from . import hill_polarization
from . import utils

# from . import visualization
from . import operators
from . import fabric_tensors
from . import approximation
from . import approximation_docstrings
from . import orientation_averager



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Utilities
"""
import numpy as np
import sys


class Ex(Exception):
    """Exception wrapping all exceptions of this package"""

    pass


def isinvertible(matrix):
    return np.linalg.cond(matrix) < 1 / sys.float_info.epsilon



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Polarizations for isotropic matrices
"""

import numpy as np

from mechmean import utils
import mechkit

TOLERANCE_SPHERE = 1e-4


class PConverter(object):
    def __init__(self):
        self.con = mechkit.notation.Converter()

    def eshelby_to_hill(self, E, matrix):
        E_m = self.con.to_mandel6(E)
        C_m = self.con.to_mandel6(matrix.stiffness)
        P = E_m @ np.linalg.inv(C_m)
        return P


class Factory(object):
    """Factory layer for Hill polarization tensors based on
    :any:`mechmean.hill_polarization.Castaneda`

    Alternative implementations based on other references are given in
    :any:`mechmean.hill_polarization_alternatives`
    and tested against this factory.
    """

    def spheroid(self, aspect_ratio, matrix):
        """
        Parameters
        ----------
        aspect_ratio : float
            Ratio of first half axis and remaining half axes of spheroid
        matrix : `mechkit.material.Isotropic`
            Isotropic matrix material

        Returns
        -------
        np.array (mandel6_4)
            Hill polarization tensor
        """
        return Castaneda().spheroid(aspect_ratio=aspect_ratio, matrix=matrix)

    def sphere(self, matrix):
        """
        Parameters
        ----------
        matrix : `mechkit.material.Isotropic`
            Isotropic matrix material

        Returns
        -------
        np.array (mandel6_4)
            Hill polarization tensor
        """
        return Castaneda().sphere(matrix=matrix)

    def needle(self, matrix):
        """
        Parameters
        ----------
        matrix : `mechkit.material.Isotropic`
            Isotropic matrix material

        Returns
        -------
        np.array (mandel6_4)
            Hill polarization tensor
        """
        return Castaneda().needle(matrix=matrix)


class Castaneda(object):
    r"""Hill polarization tensor"""

    def __init__(self):
        self.tolerance = TOLERANCE_SPHERE

        self.con = mechkit.notation.Converter()
        self.tensors = mechkit.tensors.Basic()

    def needle(self, matrix):
        r"""[Kehrer2019]_ formula (A.7)"""
        la = matrix.la
        mu = matrix.G

        k = 1 / 4.0 * 1.0 / (la + 2.0 * mu)
        m = 0.0
        r = 0.0
        p = 1 / 2.0 * (la + 3.0 * mu) / (4.0 * (la + 2.0 * mu) * mu)
        q = 1 / 2.0 * 1.0 / (4.0 * mu)

        return self._P_by_kmpqr(k=k, m=m, p=p, q=q, r=r)

    def spheroid(self, aspect_ratio, matrix):
        """Combination of
        appendix C of [Castaneda1997]_,
        equation (4.39) [Willis1981]_,
        formulas (2.57) and (2.58) in [Brylka2017]_.

        The first principle semi-axis is aligned with the global axis e_0.
        """

        kwargs = locals()
        kwargs.pop("self")

        func = self._get_func_oblate_prolate_or_sphere(aspect_ratio=aspect_ratio)
        return func(**kwargs)

    def _get_func_oblate_prolate_or_sphere(self, aspect_ratio):

        tol = self.tolerance

        if aspect_ratio <= 1.0 - tol:
            f = self._oblate_or_prolate
        elif (1.0 - tol < aspect_ratio) and (aspect_ratio < 1.0 + tol):
            f = self.sphere
        else:
            f = self._oblate_or_prolate
        return f

    def _oblate_or_prolate(self, aspect_ratio, matrix):
        def h(a):
            """Temporary parameter depending on aspect ratio"""
            a2 = a * a
            if a < 1.0:
                h = (a * (np.arccos(a) - a * np.sqrt(1.0 - a2))) / np.power(
                    1.0 - a2, 3.0 / 2.0
                )
            elif a == 1.0:
                raise utils.Ex("Aspect ratio = 1.0 not supported")
            else:
                h = (a * (a * np.sqrt(a2 - 1.0) - np.arccosh(a))) / np.power(
                    a2 - 1.0, 3.0 / 2.0
                )
            return h

        # Abbreviations
        a = aspect_ratio
        K = matrix.K
        G = matrix.G

        a2 = a * a
        term = (1.0 - a2) * G * (4.0 * G + 3.0 * K)

        # Calc
        h = h(a)

        m = ((G + 3.0 * K) * (2.0 * a2 - h - 2.0 * a2 * h)) / (4.0 * term)

        k = (
            G * (7.0 * h - 2.0 * a2 - 4.0 * a2 * h)
            + 3.0 * K * (h - 2.0 * a2 + 2.0 * a2 * h)
        ) / (8.0 * term)

        r = (
            G * (6.0 - 5.0 * h - 8.0 * a2 + 8.0 * a2 * h)
            + 3.0 * K * (h - 2.0 * a2 + 2.0 * a2 * h)
        ) / (2.0 * term)

        p = (
            G * (15.0 * h - 2.0 * a2 - 12.0 * a2 * h) + 3.0 * K * (3.0 * h - 2.0 * a2)
        ) / (16.0 * term)

        q = (
            2.0 * G * (4.0 - 3.0 * h - 2.0 * a2)
            + 3.0 * K * (2.0 - 3.0 * h + 2.0 * a2 - 3.0 * a2 * h)
        ) / (8.0 * term)

        return self._P_by_kmpqr(k=k, m=m, p=p, q=q, r=r)

    def sphere(self, matrix, **kwargs):
        """[Willis1981]_ eq. (4.39)"""
        K = matrix.K
        G = matrix.G

        alpha = 1.0 / (3.0 * K + 4.0 * G)
        beta = (3.0 * (K + 2.0 * G)) / (5.0 * G * (3.0 * K + 4.0 * G))

        H = alpha * self.tensors.P1 + beta * self.tensors.P2

        return self.con.to_mandel6(H)

    def _P_by_kmpqr(self, k, m, p, q, r):
        P = np.array(
            [
                [r, m, m, 0.0, 0.0, 0.0],
                [m, k + p, k - p, 0.0, 0.0, 0.0],
                [m, k - p, k + p, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 2.0 * p, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 2.0 * q, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 2.0 * q],
            ],
            dtype="float64",
        )
        return P



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import mechkit

import mechmean


def test_fabric_tensor_first_kind_discrete():
    """Compare einsum-implementation with loop-implementation"""

    orientations = np.random.rand(10, 3)  # Ten random vectors in 3D

    # Normalize orientations
    orientations = [np.array(v) / np.linalg.norm(v) for v in orientations]

    # Symmetrize orientations
    orientations_reversed = [-v for v in orientations]
    orientations = orientations + orientations_reversed

    def oT_loops(orientations, order=4):
        N = np.zeros((3,) * order)
        for p in orientations:
            out = p
            for index in range(order - 1):
                out = np.multiply.outer(out, p)
            N[:] = N[:] + out
        N = N / len(orientations)
        return N

    for order in range(1, 10):
        assert np.allclose(
            mechmean.fabric_tensors.first_kind_discrete(
                order=order,
                orientations=orientations,
            ),
            oT_loops(
                order=order,
                orientations=orientations,
            ),
        )


def test_fabric_tensor_first_kind_discrete_benchmarks():
    """Compare with known benchmarks from presentation M. Schneider 26.09.2018"""

    orientations = [
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 0.0, 1.0],
    ]

    N2_ref = np.zeros((3, 3))
    N2_ref[0, 0] = N2_ref[1, 1] = N2_ref[2, 2] = 1.0 / 3.0

    N4_ref = np.zeros((6, 6))
    N4_ref[0, 0] = N4_ref[1, 1] = N4_ref[2, 2] = 1.0 / 3.0
    converter = mechkit.notation.Converter()
    N4_ref_t = converter.to_tensor(N4_ref)

    f = mechmean.fabric_tensors.first_kind_discrete

    assert np.allclose(
        N2_ref,
        f(
            order=2,
            orientations=orientations,
        ),
    )

    assert np.allclose(
        N4_ref_t,
        f(
            order=4,
            orientations=orientations,
        ),
    )


# ##########################################################################
# ##########################    Trivial tests     ##########################
# ##########################################################################


def test_N2_iso():
    """This test is trivial. The contraction is a
    matrix-vector multiplication with vector [1, 1, 1, 0, 0, 0]
    which results in 1/5 * (3/3 + 1/3 + 1/3) = 1/3
    """

    con = mechkit.notation.Converter()

    N4_iso = (
        1.0
        / 5.0
        * np.array(
            [
                [1.0, 1.0 / 3.0, 1.0 / 3.0, 0.0, 541.0, 0.0],
                [1.0 / 3.0, 1.0, 1.0 / 3.0, 0.0, 0.0, 0.0],
                [1.0 / 3.0, 1.0 / 3.0, 1.0, 6874.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 2.0 / 3.0, 0.0, 34345.0],
                [0.0, 0.0, 0.0, 34.0, 3434.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 2.0 / 3.0],
            ]
        )
    )

    N2_iso = 1.0 / 3.0 * np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])

    contracted = np.einsum(
        "ijkl, kl->ij",
        con.to_tensor(N4_iso),
        mechkit.tensors.Basic().I2,
    )

    assert np.allclose(N2_iso, contracted)


def test_N2_planar_iso():
    """This test is trivial. The contraction is a
    matrix-vector multiplication with vector [1, 1, 1, 0, 0, 0]
    """

    con = mechkit.notation.Converter()

    N4_planar_iso = (
        1.0
        / 8.0
        * np.array(
            [
                [3.0, 1.0, 0.0, 0.0, 45333.0, 0.0],
                [1.0, 3.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 543.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 453354.0, 0.0],
                [0.0, 0.0, 0.0, 453453.0, 0.0, 2.0],
            ]
        )
    )

    N2_planar_iso = np.array([[0.5, 0.0, 0.0], [0.0, 0.5, 0.0], [0.0, 0.0, 0.0]])

    contracted = np.einsum(
        "ijkl, kl->ij",
        con.to_tensor(N4_planar_iso),
        mechkit.tensors.Basic().I2,
    )

    assert np.allclose(N2_planar_iso, contracted)


if __name__ == "__main__":
    test_N2_planar_iso()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Polarizations for isotropic matrices
"""

import numpy as np
from scipy.spatial.transform import Rotation

import mechkit

TOLERANCE_SPHERE = 1e-4


class PConverter(object):
    def __init__(self):
        self.con = mechkit.notation.Converter()

    def eshelby_to_hill(self, E, matrix):
        E_m = self.con.to_mandel6(E)
        C_m = self.con.to_mandel6(matrix.stiffness)
        P = E_m @ np.linalg.inv(C_m)
        return P


class Mura(object):
    """Hill polarization tensor following [Mura1987]_.
    Detailed references are available in source code docstrings"""

    def __init__(self):
        self.tolerance = TOLERANCE_SPHERE

    def spheroid(self, **kwargs):
        """First axis is aligned in e_0 direction"""

        E = self._spheroid_eshelby(**kwargs)
        # Convert Eshelby to Hill
        H = PConverter().eshelby_to_hill(E=E, matrix=kwargs["matrix"])
        return H

    def _spheroid_eshelby(self, aspect_ratio, matrix):
        """Calc Eshelby polarization tensor
        First axis is aligned in e_0 direction
        """

        axes = self._axes_by_aspect_ratio(aspect_ratio=aspect_ratio)

        function_I = self._get_function_I(aspect_ratio=aspect_ratio)

        integral = function_I(axes=axes)

        E = self._eshelby_by_integrals(
            a=axes, I=integral, poisson_matrix=matrix.poisson
        )
        return E

    def _int_index_by_str_add(self, *args):
        """String-add arguments to integer index"""
        return int("".join(map(str, args)))

    def _axes_by_aspect_ratio(self, aspect_ratio):
        """Calc principal half axes of spheroid with a[1] != a[2] == a[3] == 1.0

        Returns
        -------
        dict
            Principal half axes.
        """

        a = {}
        a[2] = a[3] = 1.0
        a[1] = aspect_ratio
        return a

    def _get_function_I(self, aspect_ratio):
        """Select function by aspect_ratio"""

        tol = self.tolerance

        if aspect_ratio <= 1.0 - tol:
            f = self._I_oblate_spheroid
        elif (1.0 - tol < aspect_ratio) and (aspect_ratio < 1.0 + tol):
            f = self._I_sphere
        else:
            f = self._I_prolate_spheroid
        return f

    def _I_sphere(self, axes):
        """Calc integrals for sphere following [Mura1987]_ page 79, (11.21)

        Parameters
        ----------
        axes : dict
            Principal half axes.
            Required keys are [1, 2, 3].

        Returns
        -------
        dict
            Keys are [1, 2, 3, 11, 22, 33, 12, 21, 23, 32, 31, 13].
        """

        a = axes

        tol = self.tolerance
        assert all(
            [
                1.0 - tol < a[1] / a[2],
                a[1] / a[2] < 1.0 + tol,
                1.0 - tol < a[1] / a[3],
                a[1] / a[3] < 1.0 + tol,
            ]
        ), "Aspect ratio of sphere must be 1.0"

        r = a[1]

        I = {}
        I[1] = I[2] = I[3] = 4.0 * np.pi / 3.0
        I[11] = I[22] = I[33] = I[12] = I[23] = I[31] = 4.0 * np.pi / (5.0 * r * r)

        I[21] = I[12]
        I[13] = I[31]
        I[32] = I[23]
        return I

    def _I_oblate_spheroid(self, axes):
        """Calc integrals of oblate spheroid [Mura1987]_ page 84, (11.28)

        Parameters
        ----------
        axes : dict
            Principal half axes.
            Required keys are [1, 2, 3].

        Returns
        -------
        dict
            Keys are [1, 2, 3, 11, 22, 33, 12, 21, 23, 32, 31, 13].
        """

        a = axes

        assert all(
            [a[1] < a[2], a[2] == a[3]]
        ), "Aspect ratio of oblate spheroid must be less than 1.0"

        I = {}
        I[3] = I[2] = (
            (2.0 * np.pi * a[3] * a[3] * a[1])
            / (np.power(a[3] * a[3] - a[1] * a[1], 3.0 / 2.0))
            * (
                np.arccos(a[1] / a[3])
                - a[1] / a[3] * np.power(1.0 - a[1] * a[1] / (a[3] * a[3]), 0.5)
            )
        )

        I[1] = 4.0 * np.pi - 2.0 * I[3]

        I[33] = I[22] = I[32] = np.pi / (a[3] * a[3]) - (I[3] - I[1]) / (
            4.0 * (a[1] * a[1] - a[3] * a[3])
        )

        I[31] = I[21] = (I[3] - I[1]) / (a[1] * a[1] - a[3] * a[3])

        I[11] = (4.0 * np.pi) / (3.0 * a[1] * a[1]) - 2.0 / 3.0 * I[31]

        I[12] = I[21]
        I[13] = I[31]
        I[23] = I[32]
        return I

    def _I_prolate_spheroid(self, axes):
        """Calc integrals of prolate spheroid [Mura1987]_, page 84, (11.29)

        Parameters
        ----------
        axes : dict
            Principal half axes.
            Required keys are [1, 2, 3].

        Returns
        -------
        dict
            Keys are [1, 2, 3, 11, 22, 33, 12, 21, 23, 32, 31, 13].
        """

        a = axes

        assert all(
            [a[1] > a[2], a[2] == a[3]]
        ), "Aspect ratio of prolate spheroid must be greater than 1.0"

        I = {}
        I[2] = I[3] = (
            (2.0 * np.pi * a[1] * a[3] * a[3])
            / (np.power(a[1] * a[1] - a[3] * a[3], 3.0 / 2.0))
            * (
                a[1] / a[3] * np.power(a[1] * a[1] / (a[3] * a[3]) - 1.0, 0.5)
                - np.arccosh(a[1] / a[3])
            )
        )

        I[1] = 4.0 * np.pi - 2.0 * I[2]

        I[12] = I[13] = (I[2] - I[1]) / (a[1] * a[1] - a[2] * a[2])

        I[11] = (4.0 * np.pi) / (3.0 * a[1] * a[1]) - 2.0 / 3.0 * I[12]

        I[22] = I[33] = I[23] = np.pi / (a[2] * a[2]) - (I[2] - I[1]) / (
            4.0 * (a[1] * a[1] - a[2] * a[2])
        )

        I[21] = I[12]
        I[31] = I[13]
        I[32] = I[23]
        return I

    def _eshelby_by_integrals(self, a, I, poisson_matrix):
        """Create Eshelby polarization tensor from non-vanishing integrals
        [Mura1987]_, page 77

        Parameters
        ----------
        a : dict
            Principal half axes.
            Required keys are [1, 2, 3].

        I : dict
            Required keys are [1, 2, 3, 11, 22, 33, 12, 21, 23, 32, 31, 13].

        poisson_matrix : float
            Poisson ratio of the matrix.

        Returns
        -------
        np.array
            Eshelby polarization tensor.
            Symmetries: Minor left and minor right. No major symmetry.
        """

        # Aliases
        add = self._int_index_by_str_add

        # Calc recurring terms once
        t1 = 8.0 * np.pi * (1.0 - poisson_matrix)
        t2 = 1.0 - 2.0 * poisson_matrix
        t3 = t2 / t1

        # Start with zeros
        E = np.zeros((3, 3, 3, 3), dtype="float64")

        simultaneous_cyclic_permutations = [
            [1, 2, 3],
            [2, 3, 1],
            [3, 1, 2],
        ]

        # Define components of polarization tensor by integrals
        for k, l, m in simultaneous_cyclic_permutations:
            k0, l0, m0 = k - 1, l - 1, m - 1
            E[k0, k0, k0, k0] = 3.0 / t1 * a[k] * a[k] * I[add(k, k)] + t3 * I[k]
            E[k0, k0, l0, l0] = 1.0 / t1 * a[l] * a[l] * I[add(k, l)] - t3 * I[k]
            E[k0, k0, m0, m0] = 1.0 / t1 * a[m] * a[m] * I[add(k, m)] - t3 * I[k]
            E[k0, l0, k0, l0] = E[l0, k0, k0, l0] = E[k0, l0, l0, k0] = E[
                l0, k0, l0, k0
            ] = (a[k] * a[k] + a[l] * a[l]) / 2.0 * 1.0 / t1 * I[
                add(k, l)
            ] + t3 / 2.0 * (
                I[k] + I[l]
            )
        return E


class Ortolano(object):
    """Hill polarization tensor [Ortolano2013]_, [Friebel2007]_"""

    def __init__(self):
        self.con = mechkit.notation.Converter()

    def needle(self, matrix):
        """Spheroid with an infite aspect_ratio

        Cylinders of original formulas point into z-direction.
        As the x-direction is the common symmetry axis,
        Eshelbys tensor is rotated to point into x-direction.
        """
        nu = matrix.poisson

        S = np.zeros((3, 3, 3, 3), dtype="float64")
        term1 = 8.0 * (1.0 - nu)

        S[0, 0, 0, 0] = S[1, 1, 1, 1] = (5.0 - 4.0 * nu) / term1
        S[0, 0, 1, 1] = S[1, 1, 0, 0] = (4.0 * nu - 1.0) / term1
        S[0, 0, 2, 2] = S[1, 1, 2, 2] = nu / (2.0 * (1.0 - nu))
        # Bugfix in next line: S[0, 1, 1, 0] instead of two times S[1, 0, 0, 1]
        S[0, 1, 0, 1] = S[1, 0, 0, 1] = S[0, 1, 1, 0] = S[1, 0, 1, 0] = (
            3.0 - 4.0 * nu
        ) / term1
        S[1, 2, 1, 2] = S[2, 1, 1, 2] = S[1, 2, 2, 1] = S[2, 1, 2, 1] = 1.0 / 4.0
        S[2, 0, 2, 0] = S[0, 2, 2, 0] = S[2, 0, 0, 2] = S[0, 2, 0, 2] = 1.0 / 4.0

        # Rotate from z-direction into x-direction
        Q = Rotation.from_rotvec(np.pi / 2 * np.array([0.0, 1.0, 0.0])).as_matrix()
        S_rotated = np.einsum("ij, kl, mn, pq, jlnq -> ikmp", Q, Q, Q, Q, S)

        H = PConverter().eshelby_to_hill(E=S_rotated, matrix=matrix)
        return H



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import mechkit
import mechmean
from mechmean import material
from mechmean import orientation_averager_alternatives


def test_compare_AdvaniTucker_N_D():

    sym = mechmean.operators.sym
    con = mechkit.notation.Converter()

    for i in range(10):
        f = material.AlignedStiffnessFactory()
        B = f.positiv_definit(label="hexagonal_axis1")

        # B = np.array([[0.7327, 0.8   , 0.8   , 0.    , 0.    , 0.    ],
        #        [0.8   , 0.9987, 0.9116, 0.    , 0.    , 0.    ],
        #        [0.8   , 0.9116, 0.9987, 0.    , 0.    , 0.    ],
        #        [0.    , 0.    , 0.    , 0.1743, 0.    , 0.    ],
        #        [0.    , 0.    , 0.    , 0.    , 0.3402, 0.    ],
        #        [0.    , 0.    , 0.    , 0.    , 0.    , 0.3402]])

        rand = sym(np.random.rand(3, 3, 3, 3))
        N4 = con.to_mandel6(rand) - 1.0 / 6.0 * (
            np.einsum("ijij->", rand) - 1.0
        ) * np.eye(6, 6)

        kanatani = mechmean.fabric_tensors.KanataniFactory(N=N4)
        D2 = kanatani.D2
        D4 = kanatani.D4

        at_D = orientation_averager_alternatives.AdvaniTucker_in_kanatani_third_kind(
            D2, D4
        )
        at = mechmean.orientation_averager.AdvaniTucker(N4)

        av_D = at_D.average(B)
        av = at.average(B)

        for i in range(5):
            print("N_base_{}".format(i))
            print(con.to_mandel9(at.base[i, :, :]))
            print("D_base_{}".format(i))
            print(con.to_mandel9(at_D.base[i, :, :]))

        print("av_D")
        print(av_D)

        print("av")
        print(av)

        print("av-av_D")
        print(av - av_D)

        assert np.allclose(av, av_D)



import itertools
import numpy as np
import mechkit


def sym(tensor, sym_axes=None):
    """Symmetrize selected axes of tensor

    If no sym_axes are specified, all axes are symmetrized
    """
    base_axis = np.array(range(len(tensor.shape)))

    sym_axes = base_axis if sym_axes is None else sym_axes

    perms = itertools.permutations(sym_axes)

    axes = list()
    for perm in perms:
        axis = base_axis.copy()
        axis[sym_axes] = perm
        axes.append(axis)

    return 1.0 / len(axes) * sum(tensor.transpose(axis) for axis in axes)


def sym_inner_mandel(tensor):
    con = mechkit.notation.Converter()
    t_mandel = con.to_mandel6(tensor)
    return con.to_like(
        inp=0.5 * (t_mandel + t_mandel.transpose()),
        like=tensor,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Material related operations
"""

import numpy as np
import mechkit

tensors = mechkit.tensors.Basic()


class SpecificMaterial(object):
    def __init__(self):
        self.con = mechkit.notation.VoigtConverter(silent=True)

    def _copy_upper_triangle(self, matrix):
        r"""Copy upper triangle to lower triangle, i.e. make symmetric"""
        index_lower_triangle = np.tril_indices(6, -1)
        matrix[index_lower_triangle] = matrix.T[index_lower_triangle]
        return matrix

    def _voigt_to_tensor(self, stiffness):
        return self.con.to_tensor(
            self.con.voigt_to_mandel6(stiffness, voigt_type="stiffness")
        )


class AlignedStiffnessFactory(SpecificMaterial):
    def __init__(self):
        self.number_independent_param = {"hexagonal_axis1": 5}
        self.func = {"hexagonal_axis1": self.hexagonal_axis1}
        super().__init__()

    def positiv_definit(self, label):
        nbr_param = self.number_independent_param[label]
        func = self.func[label]

        eigen = np.array([-1])
        while not all(eigen > 0):
            C = func(*np.random.rand(nbr_param).tolist())
            eigen = np.linalg.eig(C)[0]
        return C

    def hexagonal_axis1(self, C1111, C1122, C1133, C2222, C1313):
        voigt_half = np.array(
            [
                [C1111, C1122, C1122, 0, 0, 0],
                [0, C2222, C1133, 0, 0, 0],
                [0, 0, C2222, 0, 0, 0],
                [0, 0, 0, C2222 - C1133, 0, 0],
                [0, 0, 0, 0, C1313, 0],
                [0, 0, 0, 0, 0, C1313],
            ],
            dtype="float64",
        )
        voigt = self._copy_upper_triangle(matrix=voigt_half)
        return self.con.voigt_to_mandel6(voigt, voigt_type="stiffness")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import runpy
import os
import pytest
import glob

pytestmark = pytest.mark.scripts


THIS_FILES_DIR_PATH = os.path.realpath(os.path.dirname(__file__))


def get_paths_of_scripts():
    exclude_sub_strings = ["do_not_execute"]
    plot_script_paths = glob.glob(
        os.path.join(os.path.dirname(THIS_FILES_DIR_PATH), "examples", "*.py")
    )
    plot_script_paths_sorted = sorted(plot_script_paths)
    plot_script_paths_sorted_reduced = [
        p
        for p in plot_script_paths_sorted
        if not any(sub in p for sub in exclude_sub_strings)
    ]
    return plot_script_paths_sorted_reduced


class Test_scripts:
    @pytest.mark.parametrize(
        "path_script",
        get_paths_of_scripts(),
    )
    def test_execute_scripts(self, path_script):

        print(f"Execut script:\n{path_script}")
        runpy.run_path(path_script, init_globals={}, run_name="__main__")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import mechkit
import numpy as np
import pprint
from itertools import groupby
import pytest
import functools

import mechmean

np.set_printoptions(linewidth=140, precision=2)

#########################################
# Fixtures and data


def run_test_multiple_times(func, nbr=10):
    inps = []
    inps.append(create_standard_input())
    for i in range(10):
        inps.append(create_random_inp())

    for inp in inps:
        pprint.pprint(inp)
        func(inp)


def create_random_inp():

    bounds = {
        "E_f": {"low": 10.0, "up": 1000},
        "E_m": {"low": 1.0, "up": 1000},
        "c_f": {"low": 0.0, "up": 1},
        "k": {"low": 0.0, "up": 1},
        "nu_f": {"low": 0.0, "up": 0.5},
        "nu_m": {"low": 0.0, "up": 0.5},
    }
    inp = {}
    for key, bound in bounds.items():
        inp[key] = np.random.uniform(bound["low"], bound["up"])

    inp["N4"] = N4_from_orientations(create_random_orientations())
    add_inclusion_matrix(inp)
    return inp


def add_inclusion_matrix(inp):
    inp["inclusion"] = mechkit.material.Isotropic(E=inp["E_f"], nu=inp["nu_f"])
    inp["matrix"] = mechkit.material.Isotropic(E=inp["E_m"], nu=inp["nu_m"])
    return inp


def create_random_orientations():
    bounds_nbr_orientations = {"low": 1.0, "up": 100}
    bounds_value_orientations = {"low": 0.0, "up": 1.0}

    b_nbr = bounds_nbr_orientations
    b_val = bounds_value_orientations

    orientations = []
    for i in range(int(np.random.uniform(b_nbr["low"], b_nbr["up"]))):
        orientations.append(
            [np.random.uniform(b_val["low"], b_val["up"]) for i in range(3)]
        )
    return orientations


def N4_from_orientations(orientations):
    N4 = mechmean.fabric_tensors.first_kind_discrete(order=4, orientations=orientations)
    return N4


def create_standard_input():
    inp = {
        "E_f": 73.0,
        "E_m": 3.4,
        "c_f": 0.22,
        "k": 1.0,
        "nu_f": 0.22,
        "nu_m": 0.385,
        "N4": np.array(
            [
                [4.09e-01, 1.48e-01, 1.03e-02, -2.20e-03, -1.86e-02, 3.52e-02],
                [1.48e-01, 2.51e-01, 6.50e-03, -2.00e-03, -5.50e-03, 3.11e-02],
                [1.03e-02, 6.50e-03, 9.70e-03, 8.00e-04, -1.20e-03, 4.00e-04],
                [-2.20e-03, -2.00e-03, 8.00e-04, 1.30e-02, 5.00e-04, -7.70e-03],
                [-1.86e-02, -5.50e-03, -1.20e-03, 5.00e-04, 2.06e-02, -3.20e-03],
                [3.52e-02, 3.11e-02, 4.00e-04, -7.70e-03, -3.20e-03, 2.97e-01],
            ]
        ),
    }
    add_inclusion_matrix(inp)
    return inp


def input_dict_from_input(inp, averager):
    input_dict = {
        "phases": {
            "inclusion": {
                "material": inp["inclusion"],
                "volume_fraction": inp["c_f"],
            },
            "matrix": {"material": inp["matrix"]},
        },
        "k": inp["k"],
        "averaging_func": averager.average,
    }
    return input_dict


#########################################
# Tests


def test_mori_tanaka_orientation_averaged_comparison_implementations():
    """Successful but private"""
    assert True


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_hs2step_kehrer_compare_implementations_Kehrer2019_Kehrer2019HSW():
    run_test_multiple_times(
        func=hs2step_kehrer_compare_implementations_Kehrer2019_Kehrer2019HSW,
    )


def hs2step_kehrer_compare_implementations_Kehrer2019_Kehrer2019HSW(inp):

    averager = mechmean.orientation_averager.AdvaniTucker(N4=inp["N4"])

    input_dict = input_dict_from_input(inp, averager)

    kehrer = mechmean.approximation.Kehrer2019(**input_dict)

    kehrer.calc_C_eff()

    ###########################

    hashin = mechmean.approximation.Kehrer2019HSW(**input_dict)
    hashin.calc_C_eff()

    ###########################
    # Print

    comparisons = {
        "all": {"slice": np.s_[:, :], "tolerance": {"rtol": 1e-05, "atol": 1e-04}},
    }

    comparisons = {
        "C_ud_low": {
            "pair": [kehrer._C_ud_low, hashin.C_ud_low],
            "tolerance": {"rtol": 1e-05, "atol": 1e-04},
        },
        "C_ud_upp": {
            "pair": [kehrer._C_ud_upp, hashin.C_ud_upp],
            "tolerance": {"rtol": 1e-05, "atol": 1e-04},
        },
        "C_eff_upp": {
            "pair": [kehrer.C_eff_upp, hashin.C_eff_upp],
            "tolerance": {"rtol": 1e-05, "atol": 1e-04},
        },
        "C_eff_low": {
            "pair": [kehrer.C_eff_low, hashin.C_eff_low],
            "tolerance": {"rtol": 1e-05, "atol": 1e-04},
        },
    }

    for key, comp in comparisons.items():
        kehrer, hashin = comp["pair"]
        printQueue = [
            "key",
            "kehrer-hashin",
            "np.divide(kehrer-hashin, hashin)",
        ]

        # Print
        with np.errstate(divide="ignore"):
            for val in printQueue:
                print(val)
                print(eval(val), "\n")

        assert np.allclose(kehrer, hashin, **comp["tolerance"])


def test_hashin_shtrikman_implementations():
    """Successful but private"""
    assert True


def test_MT_is_HS_with_matrix_as_reference_material():

    inp = create_standard_input()

    add_inclusion_matrix(inp)

    P_func = functools.partial(
        mechmean.hill_polarization.Castaneda().spheroid, aspect_ratio=1.0
    )

    hashin = mechmean.approximation.HSW_VolumeFraction(
        phases={
            "i": {"material": inp["inclusion"], "volume_fraction": inp["c_f"]},
            "m": {"material": inp["matrix"], "volume_fraction": 1.0 - inp["c_f"]},
        },
        P_func=P_func,
    )

    mori = mechmean.approximation.MoriTanaka(
        phases={
            "inclusion": {
                "material": inp["inclusion"],
                "hill_polarization": P_func(matrix=inp["matrix"]),
                "volume_fraction": inp["c_f"],
            },
            "matrix": {"material": inp["matrix"]},
        },
    )

    assert np.allclose(hashin.calc_C_eff(ref_material=inp["matrix"]), mori.calc_C_eff())


def test_HSW_strain_localization_sum_to_one():
    run_test_multiple_times(func=HSW_strain_localization_sum_to_one)


def HSW_strain_localization_sum_to_one(inp):

    input_dict = {
        "phases": {
            "i": {"material": inp["inclusion"], "volume_fraction": inp["c_f"]},
            "m": {"material": inp["matrix"], "volume_fraction": 1.0 - inp["c_f"]},
        },
        "P_func": functools.partial(
            mechmean.hill_polarization.Castaneda().spheroid, aspect_ratio=1.0
        ),
    }

    hashin = mechmean.approximation.HSW_VolumeFraction(**input_dict)

    A = {
        key: hashin.calc_strain_localization_tensor(
            key_phase=key, ref_material=inp["matrix"]
        )
        for key in ["i", "m"]
    }

    A_av = np.zeros((6, 6))
    for key, val in input_dict["phases"].items():
        A_av += val["volume_fraction"] * A[key]

    printQueue = [
        "A",
        "A_av",
    ]
    for val in printQueue:
        print(val)
        pprint.pprint(eval(val))
        print()

    assert np.allclose(A_av, np.eye(6))


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_compare_implementations_Kehrer2019HSW_HSW2StepInterpolatedReferenceMaterial():
    run_test_multiple_times(
        func=compare_implementations_Kehrer2019HSW_HSW2StepInterpolatedReferenceMaterial,
    )


def compare_implementations_Kehrer2019HSW_HSW2StepInterpolatedReferenceMaterial(inp):

    averager = mechmean.orientation_averager.AdvaniTucker(N4=inp["N4"])

    input_dict = input_dict_from_input(inp, averager)

    kehrerHSW = mechmean.approximation.Kehrer2019HSW(**input_dict)
    kehrerHSW.calc_C_eff()

    k = input_dict.pop("k")
    input_dict["k2"] = k

    generic_upper = mechmean.approximation.HSW2StepInterpolatedReferenceMaterial(
        **input_dict, k1=1.0
    )
    generic_upper.calc_C_eff()

    generic_lower = mechmean.approximation.HSW2StepInterpolatedReferenceMaterial(
        **input_dict, k1=0.0
    )
    generic_lower.calc_C_eff()

    ###########################
    # Print

    comparisons = {
        "all": {
            "slice": np.s_[:, :],
            "tolerance": {"rtol": 1e-05, "atol": 1e-04},
        },
    }

    comparisons = {
        "C_ud_low": {
            "pair": [kehrerHSW.C_ud_low, generic_lower.C_ud],
            "tolerance": {"rtol": 1e-05, "atol": 1e-04},
        },
        "C_ud_upp": {
            "pair": [kehrerHSW.C_ud_upp, generic_upper.C_ud],
            "tolerance": {"rtol": 1e-05, "atol": 1e-04},
        },
        "C_eff_upp": {
            "pair": [kehrerHSW.C_eff_upp, generic_upper.C_eff],
            "tolerance": {"rtol": 1e-05, "atol": 1e-04},
        },
        "C_eff_low": {
            "pair": [kehrerHSW.C_eff_low, generic_lower.C_eff],
            "tolerance": {"rtol": 1e-05, "atol": 1e-04},
        },
    }

    for key, comp in comparisons.items():
        kehrer, generic = comp["pair"]
        printQueue = [
            "key",
            "kehrer-generic",
            "np.divide(kehrer-generic, generic)",
        ]

        # Print
        with np.errstate(divide="ignore"):
            for val in printQueue:
                print(val)
                print(eval(val), "\n")

        assert np.allclose(
            kehrer,
            generic,
            **comp["tolerance"],
        )


if __name__ == "__main__":
    test_compare_implementations_Kehrer2019HSW_HSW2StepInterpolatedReferenceMaterial()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Orientation averager
"""

import numpy as np
import mechkit
import mechmean


class AdvaniTucker_in_kanatani_third_kind(mechmean.orientation_averager.AdvaniTucker):
    def __init__(self, D2, D4):
        self._con = mechkit.notation.Converter()
        self.base = self.get_base(D2, D4)

    def get_base(self, D2, D4):

        tensors = mechkit.tensors.Basic()
        I2 = tensors.I2
        I4s = tensors.I4s
        P1 = tensors.P1
        sym = mechmean.operators.sym

        D4 = self._con.to_tensor(D4)
        D2 = self._con.to_tensor(D2)

        def dyad(A, B):
            return np.einsum("ij, kl->ijkl", A, B)

        def box(A, B):
            return np.einsum("ij, kl->iklj", A, B)

        D_box_I = box(D2, I2)
        I_box_D = box(I2, D2)

        D_dyad_I = dyad(D2, I2)
        I_dyad_D = dyad(I2, D2)

        base = np.zeros((5, 3, 3, 3, 3))
        base[0, :] = 8.0 / 315.0 * D4 + 4.0 / 35.0 * sym(I_dyad_D) + 3.0 / 5.0 * sym(P1)

        base[1, :] = 2.0 / 15.0 * (D_dyad_I + I_dyad_D) + 2.0 * P1

        base[2, :] = (
            2.0
            / 15.0
            * (
                D_box_I
                + np.einsum("ijkl->ijlk", D_box_I)
                + np.einsum("ijkl->klij", I_box_D)
                + np.einsum("ijkl->ijlk", I_box_D)
            )
            + 4.0 / 3.0 * I4s
        )

        base[3, :] = dyad(I2, I2)

        base[4, :] = I4s

        return base



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import mechkit
import mechmean


class KanataniFactory(object):
    def __init__(self, N):
        self.con = mechkit.notation.Converter()
        self._I2 = mechkit.tensors.Basic().I2

        self.N = N = self.con.to_tensor(N)
        self.degree = len(N.shape)

        degrees = [x for x in range(1, self.degree + 1) if x % 2 == 0]
        for degree in reversed(degrees):
            N = self.first_kind(degree)
            setattr(self, "N{}".format(degree), N)
            setattr(self, "F{}".format(degree), self.second_kind(N))
            setattr(self, "D{}".format(degree), self.third_kind(N))

    def __getitem__(self, key):
        """Make attributes accessible dict-like."""
        return getattr(self, key)

    def first_kind(self, degree):
        nbr_times_decrease = int((self.degree - degree) / 2)
        N = self.N
        for i in range(nbr_times_decrease):
            N = self.decrease_first_kind_by_one_degree(N)
        return N

    def decrease_first_kind_by_one_degree(self, N):
        return np.einsum("...ij, ...ij->...", N, self._I2)

    def second_kind(self, N):
        degree = len(N.shape)
        func = self._get_func_second_kind(degree=degree)
        return func(N)

    def _get_func_second_kind(self, degree):
        funcs = {
            2: self.second_kind_N2,
            4: self.second_kind_N4,
        }
        return funcs[degree]

    def second_kind_N2(self, N):
        return 15.0 / 2.0 * (N - 1.0 / 5.0 * self._I2)

    def second_kind_N4(self, N):
        return (
            315.0
            / 8.0
            * (
                N
                - 2.0
                / 3.0
                * mechmean.operators.sym(
                    np.multiply.outer(self._I2, self.first_kind(degree=2))
                )
                + 1.0
                / 21.0
                * mechmean.operators.sym(np.multiply.outer(self._I2, self._I2))
            )
        )

    def third_kind(self, N):
        degree = len(N.shape)
        func = self._get_func_third_kind(degree=degree)
        return func(N)

    def _get_func_third_kind(self, degree):
        funcs = {2: self.third_kind_N2, 4: self.third_kind_N4}
        return funcs[degree]

    def third_kind_N2(self, N):
        return 15.0 / 2.0 * (N - 1.0 / 3.0 * self._I2)

    def third_kind_N4(self, N):
        return (
            315.0
            / 8.0
            * (
                N
                - 6.0
                / 7.0
                * mechmean.operators.sym(
                    np.multiply.outer(self._I2, self.first_kind(degree=2))
                )
                + 3.0
                / 35.0
                * mechmean.operators.sym(np.multiply.outer(self._I2, self._I2))
            )
        )


def evenly_distributed_vectors_on_sphere(nbr_vectors=1000):
    """
    Define nbr_vectors evenly distributed vectors on a sphere

    Using the golden spiral method kindly provided by
    stackoverflow-user "CR Drost"
    https://stackoverflow.com/a/44164075/8935243
    """
    from numpy import pi, cos, sin, arccos, arange

    indices = arange(0, nbr_vectors, dtype=float) + 0.5

    phi = arccos(1 - 2 * indices / nbr_vectors)
    theta = pi * (1 + 5 ** 0.5) * indices

    x, y, z = cos(theta) * sin(phi), sin(theta) * sin(phi), cos(phi)
    orientations = np.column_stack((x, y, z))
    return orientations


def first_kind_discrete(orientations, order=4):
    """
    Calc orientation tensors of ... kind
    """
    # Normalize orientations
    orientations = [np.array(v) / np.linalg.norm(v) for v in orientations]

    # Symmetrize orientations
    #    orientations_reversed = [-v for v in orientations]
    #    orientations = orientations + orientations_reversed

    einsumStrings = {
        1: "ij             -> j",
        2: "ij, ik         -> jk",
        3: "ij, ik, il     -> jkl",
        4: "ij, ik, il, im -> jklm",
        5: "ij, ik, il, im, in     -> jklmn",
        6: "ij, ik, il, im, in, ip -> jklmnp",
    }

    if order > 6:
        einsumStrings[order] = einsum_str_fabric_tensor_first_kind_discrete(order=order)

    einsumArgs = [orientations for i in range(order)]

    N = 1.0 / len(orientations) * np.einsum(einsumStrings[order], *einsumArgs)
    return N


def einsum_str_fabric_tensor_first_kind_discrete(order):
    """
    Generalize to higher orders:

    N = sum_i 'order'-times_dyad_product(vector)
    =
    1:  'ij             -> j',
    2:  'ij, ik         -> jk',
    3:  'ij, ik, il     -> jkl',
    4:  'ij, ik, il, im -> jklm',
    5:  'ij, ik, il, im, in     -> jklmn',
    6:  'ij, ik, il, im, in, ip -> jklmnp',
    ...
    """

    # Get list of all available characters
    import string

    letters = list(string.ascii_letters)
    letters.remove("i")

    # Create einsum string and arguments
    einsumInput = ",".join(["i" + letters[index] for index in range(order)])
    einsumOut = "".join(letters[0:order])
    einsumString = einsumInput + "->" + einsumOut

    return einsumString



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Docstrings for classes and methods in approximation.py
"""
import mechmean
import mechmean.approximation as approx

docstrings = {
    approx.TwoPhaseComposite: r"""
        Base class for approximations of two-phase materials.

        Formulations are based on average **strain localization tensors**
        :math:\mathbb{A}_{\text{j}}`
        mapping the average strain to the average strain in phase j,
        i.e.

        .. math::
            \begin{align*}
                \left<\mathbb{\varepsilon}\right>_{\text{j}}
                =
                \mathbb{A}_{\text{j}}
                \left[
                \left<\mathbb{\varepsilon}\right>
                \right]
            \end{align*}

        """,
    approx.TwoPhaseComposite.calc_A_SI_from_hill_polarization: r"""
        Calc strain localization tensor in single inclusion problem.

        .. math::
            \begin{align*}
                \mathbb{A}_{\text{i}}^{\text{SI}}
                =
                \left(
                    \mathbb{P}_{\text{i}}
                    \left(
                        \mathbb{C}_{\text{i}}
                        - \mathbb{C}_{\text{m}}
                    \right)
                    + \mathbb{I}^{\text{S}}
                \right)^{-1}
            \end{align*}

        Parameters
        ----------
        P_i : np.array (mandel6_4)
            Hill polarization
        C_i : np.array (mandel6_4)
            Stiffness of inclusion
        C_m : np.array (mandel6_4)
            Stiffness of matrix

        Returns
        -------
        np.array (mandel6_4)
            Strain localization in single inclusion problem.

        """,
    approx.TwoPhaseComposite.calc_C_eff_by_A_i: r"""
        Calc effective stiffness of two phase material

        .. math::
            \begin{align*}
                \mathbb{C}_{\text{eff}}
                =
                \mathbb{C}_{\text{m}}
                +
                c_{\text{i}}
                \left(
                    \mathbb{C}_{\text{i}}
                    - \mathbb{C}_{\text{m}}
                \right)
                \mathbb{A}_{\text{i}}^{\text{Approximated}}
            \end{align*}

        Parameters
        ----------
        c_i : float
            Volume fraction of inclusion.
        A_i : np.array (mandel6_4)
            Strain localization of inclusion.
        C_i : np.array (mandel6_4)
            Stiffness of inclusion.
        C_m : np.array (mandel6_4)
            Stiffness of matrix.

        Returns
        -------
        np.array (mandel6_4)
            Effective stiffness.

        """,
    #############################################################
    approx.MoriTanaka: r"""
        Approximate strain localization following [Mori1973]_.

        The ansatz [Gross2016]_ (equation 8.101)

        .. math::
                \begin{align*}
                    \left<\mathbb{\varepsilon}\right>_{\text{i}}
                    =
                    \mathbb{A}^{\text{SI}}
                    \left[
                        \left< \mathbb{\varepsilon} \right>_{\text{m}}
                    \right]
                \end{align*}

        leads to

        .. math::
                \begin{align*}
                    \left<\mathbb{\varepsilon}\right>
                    &=
                    c_{\text{m}}
                    \left<\mathbb{\varepsilon}\right>_{\text{m}}
                    +
                    c_{\text{i}}
                    \left<\mathbb{\varepsilon}\right>_{\text{i}}    \\
                    &=
                    c_{\text{m}}
                    \left(\mathbb{A}^{\text{SI}}\right)^{-1}
                    \left<\mathbb{\varepsilon}\right>_{\text{i}}
                    +
                    c_{\text{i}}
                    \left<\mathbb{\varepsilon}\right>_{\text{i}}    \\
                    &=
                    \underbrace{
                        \left(
                            c_{\text{m}}
                            \left(\mathbb{A}^{\text{SI}}\right)^{-1}
                            +
                            c_{\text{i}}
                            \mathbb{I}^{\text{S}}
                        \right)
                    }_{\left( \mathbb{A}_{\text{i}}^{\text{MT}} \right)^{-1}}
                    \left<\mathbb{\varepsilon}\right>_{\text{i}}
                \end{align*}

        .

        Note
        ----
            The Hill polarization of the inclusion
            :math:`\mathbb{P}_{\text{i}}`
            represents the geometry of the inclusion and depends on the
            material properties of the matrix.

        """,
    approx.MoriTanaka.__init__: r"""
        Parameters
        ----------
        phases : dict
            Valid phases are 'inclusion' and 'matrix'.
        phases['inclusion']['material'] : mechkit.material.Isotropic
            Inclusion material
        phases['inclusion']['volume_fraction'] : float
            Volume fraction of inclusion
        phases['inclusion']['hill_polarization'] : np.array (mandel6_4)
            Hill polarization
        phases['matrix']['material'] : mechkit.material.Isotropic
            Matrix material
        """,
    approx.MoriTanaka.calc_A_MT_i: r"""
        Calc strain localization by Mori-Tanaka assumption with

        .. math::
            \begin{align*}
                \mathbb{A}_{\text{i}}^{\text{MT}}
                &=
                \left(
                    c_{\text{m}}
                    \left(
                        \mathbb{A}_{\text{i}}^{\text{SI}}
                    \right)^{-1}
                    +
                    c_{\text{i}}
                    \mathbb{I}^{\text{S}}
                \right)^{-1}
            \end{align*}

        Note
        ----
            See [Weng1990]_ (2.22) for a connection between
            average strain localization tensors of the Mori-Tanaka scheme
            and Hashin-Shtrikman-Walpole scheme, as the above equation can be
            cast into the following form:

            .. math::
                \begin{align*}
                    \mathbb{A}_{\text{i}}^{\text{MT}}
                    &=
                    \mathbb{A}_{\text{i}}^{\text{SI}}
                    \left(
                        c_{\text{m}}
                        \mathbb{I}^{\text{S}}
                        +
                        c_{\text{i}}
                        \mathbb{A}_{\text{i}}^{\text{SI}}
                    \right)^{-1}    \\
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    &=
                    \mathbb{A}_{\text{i}}^{\text{SI}}
                    \left(
                        \left<
                            \mathbb{A}^{\text{SI}}
                        \right>
                    \right)^{-1}  \;\;
                        \text{with
                        $\mathbb{A}_{\text{m}}^{\text{SI}}=\mathbb{I}$
                        }      \\
                \end{align*}

            Note that
            :math:`\mathbb{A}_{\text{i}}^{\text{SI}}
            (P, C_{\text{i}}, C_{\text{surrounding material}}
            )`
            is a function of the matrix stiffness in the Mori-Tanaka context
            and a function of the stiffness of the reference material in the
            Hashin-Shtrikman context.

        Parameters
        ----------
        c_i : float
            Volume fraction of inclusion.
        A_SI_i : np.array (mandel6_4)
            Strain localization of inclusion.

        Returns
        -------
        np.array (mandel6_4)
            Strain localization Mori-Tanaka.

        """,
    approx.MoriTanaka.calc_C_eff: r"""
        Calc, set as attribute and return: Effective stiffness

        Returns
        -------
        np.array (mandel6_4)
            Effective stiffness

        """,
    #############################################################
    approx.MoriTanakaOrientationAveragedBenveniste: r"""
        Include fiber orientation in Mori-Tanaka approximation
        following [Benveniste1987]_ and [Brylka2017]_ equation (2.73).

        Use
        :math:`\left<\mathbb{A}_{\text{i}}^{\text{MT}}\right>_{\text{f}}`
        as
        :math:`\mathbb{A}_{\text{i}}^{\text{Approximated}}`
        with

        .. math::
                \begin{align*}
                    \left<\mathbb{A}_{\text{i}}^{\text{MT}}\right>_{\text{f}}
                    =
                    \left(
                        c_{\text{m}}
                        \left(
                            \left<
                                \mathbb{A}_{\text{i}}^{\text{SI}}
                            \right>_{\text{f}}
                        \right)^{-1}
                        +
                        c_{\text{i}}
                        \mathbb{I}^{\text{S}}
                    \right)^{-1}
                \end{align*}

        and with the orientation averaged strain localization of the
        single inclusion problem
        :math:`\left<\mathbb{A}_{\text{i}}^{\text{SI}}\right>_{\text{f}}`.

        Examples
        --------
        >>> import mechkit
        >>> inp = {
                'E_f': 73.0,
                'E_m': 3.4,
                'N4': np.array(
                  [[ 4.09e-01,  1.48e-01,  1.03e-02, -2.20e-03, -1.86e-02,  3.52e-02],
                   [ 1.48e-01,  2.51e-01,  6.50e-03, -2.00e-03, -5.50e-03,  3.11e-02],
                   [ 1.03e-02,  6.50e-03,  9.70e-03,  8.00e-04, -1.20e-03,  4.00e-04],
                   [-2.20e-03, -2.00e-03,  8.00e-04,  1.30e-02,  5.00e-04, -7.70e-03],
                   [-1.86e-02, -5.50e-03, -1.20e-03,  5.00e-04,  2.06e-02, -3.20e-03],
                   [ 3.52e-02,  3.11e-02,  4.00e-04, -7.70e-03, -3.20e-03,  2.97e-01]]
                  ),
                'c_f': 0.22,
                'nu_f': 0.22,
                'nu_m': 0.385,
                }

        >>> inclusion = mechkit.material.Isotropic(
                    E=inp['E_f'],
                    nu=inp['nu_f'],
                    )
        >>> matrix = mechkit.material.Isotropic(
                    E=inp['E_m'],
                    nu=inp['nu_m'],
                    )
        >>> averager = mechmean.orientation_averager.AdvaniTucker(N4=inp['N4'])
        >>> P_func = mechmean.hill_polarization.Castaneda().needle

        >>> input_dict = {
                'phases': {
                    'inclusion': {
                        'material': inclusion,
                        'volume_fraction': inp['c_f'],
                        'hill_polarization': P_func(matrix=matrix),
                        },
                    'matrix': {
                        'material': matrix,
                        },
                    },
                'averaging_func': averager.average,
                }

        >>> mori = mechmean.approximation.MoriTanakaOrientationAveragedBenveniste(**input_dict)
        >>> C_eff = mori.calc_C_eff()
        [[ 1.479e+01,  7.000e+00,  5.045e+00, -3.118e-02, -2.842e-01,  4.863e-01],
         [ 7.000e+00,  1.251e+01,  5.057e+00, -2.973e-02, -7.226e-02,  4.390e-01],
         [ 5.045e+00,  5.057e+00,  8.827e+00,  1.472e-02, -1.303e-02, -2.303e-02],
         [-3.118e-02, -2.973e-02,  1.472e-02,  3.824e+00,  1.668e-02, -1.239e-01],
         [-2.842e-01, -7.226e-02, -1.303e-02,  1.668e-02,  3.982e+00, -4.843e-02],
         [ 4.863e-01,  4.390e-01, -2.303e-02, -1.239e-01, -4.843e-02,  8.386e+00]])

        """,
    approx.MoriTanakaOrientationAveragedBenveniste.__init__: r"""
        Parameters
        ----------
        ....
            Parameters of
            :any:`mechmean.approximation.MoriTanaka`
            and in addition:
        averaging_func : function
            Function averaging fourth order tensor.

        """,
    approx.MoriTanakaOrientationAveragedBenveniste.calc_A_SI_i: r"""
        Some orientation averaging schemes are restricted to quantities
        with transversal isotropic material symmetry.
        :math:`\mathbb{A}_{\text{i}}^{\text{SI}}`
        may lack this symmetry.
        The orientation averaging scheme is applied to an
        intermediate, stiffness-like quantity
        :math:`\mathbb{A}^{\star}`
        leading to

        .. math::
               \begin{align*}
                   \mathbb{A}^{\star}
                   &=
                   \text{d}\mathbb{C}
                   \mathbb{A}^{\text{SI}}           \\
                   \left<\mathbb{A}^{\star}\right>_{\text{f}}
                   &=
                   \text{Orientation average}\left(\mathbb{A}^{\star}\right) \\
                   \left<\mathbb{A}_{\text{i}}^{\text{SI}}\right>_{\text{f}}
                   &=
                   \left(\text{d}\mathbb{C}\right)^{-1}
                   \left<\mathbb{A}^{\star}\right>_{\text{f}}   \\
                   \text{with } \;\;
                   \text{d}\mathbb{C}
                   &=
                   \mathbb{C}_{\text{i}} - \mathbb{C}_{\text{m}}.
               \end{align*}

        """,
    #############################################################
    approx.Kehrer2019: r"""
        Two-step Hashin-Shtrikman homogenization scheme for fiber reinforced
        composites [Kehrer2019]_

        **Step 1**

        .. math::
                \begin{align*}
                    \mathbb{C}_{\text{upp}}
                    &=
                    \mathbb{C}_{\text{i}}
                    +
                    c_{\text{m}}
                    \text{d} \mathbb{C}
                    \left(
                        \mathbb{I}^{\text{S}}
                        -
                        c_{\text{i}}
                        \mathbb{P}_{\text{i}}^{\text{UD}}
                        \text{d} \mathbb{C}
                    \right)^{-1}        \\
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    \mathbb{C}_{\text{low}}
                    &=
                    \mathbb{C}_{\text{m}}
                    -
                    c_{\text{i}}
                    \text{d} \mathbb{C}
                    \left(
                        \mathbb{I}^{\text{S}}
                        -
                        c_{\text{m}}
                        \mathbb{P}_{\text{m}}^{\text{UD}}
                        \text{d} \mathbb{C}
                    \right)^{-1}        \\
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    \text{d}\mathbb{C}
                    &=
                    \mathbb{C}_{\text{m}}-\mathbb{C}_{\text{i}}
                \end{align*}

        with :math:`\text{j} \in [\text{i},\text{m}]` and

            - :math:`c_{\text{j}}` : Volume fraction of phase j
            - :math:`\mathbb{C}_{\text{j}}` : Stiffness of phase j
            - :math:`\mathbb{P}_{\text{j}}^{\text{UD}}` : Hill polarization of needle shape in material of phase j

        **Step 2**

        For :math:`\text{j} \in [\text{upp},\text{low}]`

        .. math::
                \begin{align*}
                    \mathbb{C}_{\text{j}}^{\text{eff}}
                    &=
                    \mathbb{C}_{\text{0}}
                    -
                    \mathbb{P}_{\text{0}}^{-1}
                    +
                    \left<
                        \mathbb{A}_{\text{j}}^{\star}
                    \right>_{\text{orientation}}^{-1}       \\
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    \mathbb{A}_{\text{j}}^{\star}
                    &=
                    \left(
                        \mathbb{P}_{\text{0}}^{-1}
                        +
                        \mathbb{C}_{\text{j}}
                        -
                        \mathbb{C}_{\text{0}}
                    \right)^{-1}                            \\
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    \mathbb{C}_{\text{0}}
                    &=
                    \left(
                        1-k
                    \right)
                    \mathbb{C}_{\text{m}}
                    +
                    k
                    \mathbb{C}_{\text{i}}
                \end{align*}

        with

            - :math:`k` :  Scalar weight of fiber phase in reference material.
            - :math:`\mathbb{C}_{\text{upp}}^{\text{eff}}` : Upper effective stiffness
            - :math:`\mathbb{C}_{\text{low}}^{\text{eff}}` : Lower effective stiffness

        Examples
        --------
        >>> import mechkit
        >>> inp = {
                'E_f': 73.0,
                'E_m': 3.4,
                'N4': np.array(
                  [[ 4.09e-01,  1.48e-01,  1.03e-02, -2.20e-03, -1.86e-02,  3.52e-02],
                   [ 1.48e-01,  2.51e-01,  6.50e-03, -2.00e-03, -5.50e-03,  3.11e-02],
                   [ 1.03e-02,  6.50e-03,  9.70e-03,  8.00e-04, -1.20e-03,  4.00e-04],
                   [-2.20e-03, -2.00e-03,  8.00e-04,  1.30e-02,  5.00e-04, -7.70e-03],
                   [-1.86e-02, -5.50e-03, -1.20e-03,  5.00e-04,  2.06e-02, -3.20e-03],
                   [ 3.52e-02,  3.11e-02,  4.00e-04, -7.70e-03, -3.20e-03,  2.97e-01]]
                  ),
                'c_f': 0.22,
                'k': 0.5,
                'nu_f': 0.22,
                'nu_m': 0.385,
                }

        >>> inclusion = mechkit.material.Isotropic(
                    E=inp['E_f'],
                    nu=inp['nu_f'],
                    )
        >>> matrix = mechkit.material.Isotropic(
                    E=inp['E_m'],
                    nu=inp['nu_m'],
                    )
        >>> averager = mechmean.orientation_averager.AdvaniTucker(N4=inp['N4'])

        >>> input_dict = {
                'phases': {
                    'inclusion': {
                        'material': inclusion,
                        'volume_fraction': inp['c_f'],
                        },
                    'matrix': {
                        'material': matrix,
                        'volume_fraction': 1. - inp['c_f'],
                        },
                    },
                'k': inp['k'],
                'averaging_func': averager.average,
                }
        >>> hashin = mechmean.approximation.Kehrer2019(**input_dict)
        >>> C_eff = hashin.calc_C_eff()
        Effective_stiffness(
        upper=array(
        [[ 1.793e+01,  6.792e+00,  6.440e+00, -7.925e-03, -1.225e-01,  2.797e-01],
         [ 6.792e+00,  1.666e+01,  6.502e+00, -1.443e-02, -9.110e-03,  2.566e-01],
         [ 6.440e+00,  6.502e+00,  1.438e+01, -2.853e-03, -4.822e-02, -3.038e-02],
         [-7.925e-03, -1.443e-02, -2.853e-03,  8.566e+00,  1.067e-01, -7.336e-02],
         [-1.225e-01, -9.110e-03, -4.822e-02,  1.067e-01,  8.955e+00, -1.951e-02],
         [ 2.797e-01,  2.566e-01, -3.038e-02, -7.336e-02, -1.951e-02,  1.099e+01]]
        ),
        lower=array(
        [[ 1.320e+01,  6.537e+00,  5.063e+00, -2.566e-02, -2.110e-01,  4.354e-01],
         [ 6.537e+00,  1.134e+01,  5.077e+00, -2.310e-02, -5.618e-02,  3.731e-01],
         [ 5.063e+00,  5.077e+00,  8.767e+00,  8.965e-03, -8.823e-03, -1.952e-02],
         [-2.566e-02, -2.310e-02,  8.965e-03,  3.737e+00,  1.406e-02, -9.265e-02],
         [-2.110e-01, -5.618e-02, -8.823e-03,  1.406e-02,  3.847e+00, -3.933e-02],
         [ 4.354e-01,  3.731e-01, -1.952e-02, -9.265e-02, -3.933e-02,  7.122e+00]]
        )   )

        """,
    approx.Kehrer2019.__init__: r"""
        Parameters
        ----------
        phases : dict
            Valid phases are 'inclusion' and 'matrix'.
        phases['inclusion']['material'] : mechkit.material.Isotropic
            Isotropic inclusion material.
        phases['inclusion']['volume_fraction'] : float
            Volume fraction of inclusion.
        phases['matrix']['material'] : mechkit.material.Isotropic
            Isotropic matrix material.
        averaging_func : callable
            Callable averaging fourth order tensor.
        k : float in range(0,1)
            Weight of fiber phase in reference material.
        """,
    approx.Kehrer2019.calc_C_eff: r"""
        Calc, set as attribute and return: Effective stiffnesses

        Returns
        -------
        tuple(np.array, np.array) with arrays of shape (mandel6_4)
            Upper and lower effective stiffness
        """,
    #############################################################
    approx.HashinShtrikmanWalpole: r"""
        Hashin-Shtrikman scheme formulated by
        [Walpole1966_I]_, [Walpole1966_II]_
        formatted following
        [Fernandez2019]_ (equation 22)
        and applicable to non-spherical Hill polarization following
        [Walpole1969]_, [Willis1977]_ (page 190) and [Willis1981]_ (pages 36, 38)
        including a dual scheme for singular polarizations following
        [Walpole1969]_ (page 238).

        .. math::
                \begin{align*}
                    \mathbb{C}^{\text{eff}}
                    &=
                    \mathbb{C}_{\text{0}}
                    -
                    \mathbb{P}_{\text{0}}^{-1}
                    +
                    \left<
                        \mathbb{W}
                    \right>^{-1}        \\
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    \mathbb{W}
                    &=
                    \left(
                        \mathbb{C}
                        -
                        \mathbb{C}_{\text{0}}
                        +
                        \mathbb{P}_{\text{0}}^{-1}
                    \right)^{-1}
                \end{align*}

        with
            - :math:`\mathbb{C}_{\text{0}}` : Stiffness of reference material
            - :math:`\mathbb{P}_{\text{0}}` : Hill polarization of single inclusion in reference material
            - :math:`\mathbb{C}` : Stiffness at location :math:`\boldsymbol{x}` (piecewise constant phases)
            - :math:`\left<\right>` : Suitable average for given microstructure. See e.g. [Walpole1966_II] equations 24, 25

        Note
        ----
            The Hill polarization
            :math:`\mathbb{P}_{\text{0}}`
            reflects the symmetry of the two-point statistics
            (See [Willis1977]_ (page 188), [Fernandez2019]_ ).

        """,
    approx.HashinShtrikmanWalpole.__init__: r"""
        Parameters
        ----------
        phases : nested dict
            Each entry represents one phase.
        phases[....]['material'] : material with .stiffness_mandel6
            Material of phase '...'.
        phases[....]['W'] : None
            Phases must not have key 'W', because each phase-dict is
            extended by 'W'.
            The resulting extended dict is passed to averging_func.
        P_func : callcable
            Callable returning Hill polarization for given material.
        averaging_func : callable
            Callable that takes nested dict of phases and calculates
            average of 'W'.
        """,
    approx.HashinShtrikmanWalpole.calc_C_eff: r"""
        Calc effective stiffness for given reference material

        Handle singular polarizations.

        Returns
        -------
        np.array (mandel6_4)
            Effective stiffness
        """,
    approx.HashinShtrikmanWalpole.calc_strain_localization_tensor: r"""
        Calc **strain localization tensor**

        .. math::
            \begin{align*}
                \mathbb{A}_{\text{j}}
                =
                \mathbb{A}_{\text{j}}^{\text{SI}}
                \left<
                \mathbb{A}^{\text{SI}}
                \right>^{-1}
            \end{align*}

        mapping average strain to average strain in phase j.

        .. math::
            \begin{align*}
                \left<\mathbb{\varepsilon}\right>_{\text{j}}
                =
                \mathbb{A}_{\text{j}}
                \left[
                \left<\mathbb{\varepsilon}\right>
                \right]
            \end{align*}

        with

            - :math:`\mathbb{A}^{\text{SI}}` : Strain localization tensor in single inclusion problem
            - :math:`\left<\right>` : Suitable average for given microstructure
            - :math:`\left<\mathbb{\varepsilon}\right>_{\text{j}}` : Average strain in phase j
            - :math:`\left<\mathbb{\varepsilon}\right>` : Average strain

        Parameters
        ----------
        key_phase : str
            Key of phase j in dict *phases*
        ref_material : mechkit.material.Isotropic
            Reference material

        Returns
        -------
        np.array (mandel6_4)
            Strain localization
        """,
    approx.HashinShtrikmanWalpole.calc_stress_localization_tensor: r"""
        Calc **stress localization tensor**

        .. math::
            \begin{align*}
                \mathbb{B}_{\text{j}}
                =
                \mathbb{C}_{\text{j}}
                \mathbb{A}_{\text{j}}
                \mathbb{C}_{\text{0}}^{-1}
            \end{align*}

        mapping average stress to average stress in phase j.

        .. math::
            \begin{align*}
                \left<\mathbb{\sigma}\right>_{\text{j}}
                =
                \mathbb{B}_{\text{j}}
                \left[
                \left<\mathbb{\sigma}\right>
                \right]
            \end{align*}

        with

            - :math:`\mathbb{C}_{\text{j}}` : Stiffness of phase j
            - :math:`\mathbb{A}_{\text{j}}` : Strain localization tensor of phase j
            - :math:`\mathbb{C}_{\text{0}}` :  Stiffness of reference material
            - :math:`\left<\mathbb{\sigma}\right>_{\text{j}}` : Average stress in phase j
            - :math:`\left<\mathbb{\sigma}\right>` : Average stress

        Parameters
        ----------
        key_phase : str
            Key of phase j in dict *phases*
        ref_material : mechkit.material.Isotropic
            Reference material

        Returns
        -------
        np.array (mandel6_4)
            Stress localization
        """,
    #############################################################
    approx.HSW_VolumeFraction: r"""
        :any:`mechmean.approximation.HashinShtrikmanWalpole`
        with volume average calculated as weighted summation""",
    #############################################################
    approx.HSW_SinglePhaseOrientation: r"""
        :any:`mechmean.approximation.HashinShtrikmanWalpole`
        with orientation average over single phase""",
    approx.HSW_SinglePhaseOrientation.__init__: r"""
        Parameters
        ----------
        ....
            Parameters of
            :any:`mechmean.approximation.HashinShtrikmanWalpole`
            and in addition:
        orientation_averager : callable
            Returns orientation average of fourth order tensor given as first
            positional argument
        """,
    #############################################################
    approx.Kehrer2019HSW: r"""
        Implementation of :any:`mechmean.approximation.Kehrer2019` using
        :any:`mechmean.approximation.HashinShtrikmanWalpole`
        """,
    approx.Kehrer2019HSW.__init__: r"""
        See
        :any:`mechmean.approximation.Kehrer2019`
        """,
    approx.Kehrer2019HSW.calc_C_eff: r"""
        See
        :any:`mechmean.approximation.Kehrer2019`
        """,
    #############################################################
    approx.HSW2StepInterpolatedReferenceMaterial: r"""
        Based on Kehrer2019 but with interpolation parameter in both steps
        combined with the inverse scheme for singular polarizations
        leading to a generic structure
        """,
}

for identifier, docstring in docstrings.items():
    setattr(identifier, "__doc__", docstring)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Approximations of effective stiffness of two-phase materials
"""

import numpy as np
import mechmean
import mechkit
import functools
import copy
from collections import namedtuple
from numpy.linalg import inv
from mechmean import utils
import pprint


#####################################################################################
class TwoPhaseComposite(object):
    """See approximation_docstrings for docstrings"""

    def __init__(self):
        self.notation = "mandel6"
        self._I4s = mechkit.notation.Converter().to_mandel6(mechkit.tensors.Basic().I4s)

    def calc_A_SI_from_hill_polarization(self, P_i, C_i, C_m):
        return inv(P_i @ (C_i - C_m) + self._I4s)

    def calc_C_eff_by_A_i(self, c_i, A_i, C_i, C_m):
        return C_m + c_i * ((C_i - C_m) @ A_i)


#####################################################################################
class MoriTanaka(TwoPhaseComposite):
    """See approximation_docstrings for docstrings"""

    def __init__(self, phases, **kwargs):
        self._C_i = phases["inclusion"]["material"].stiffness_mandel6
        self._c_i = phases["inclusion"]["volume_fraction"]
        self._P_i = phases["inclusion"]["hill_polarization"]
        self._C_m = phases["matrix"]["material"].stiffness_mandel6
        super().__init__()

    def calc_A_MT_i(self, c_i, A_SI_i):
        return inv((1.0 - c_i) * inv(A_SI_i) + c_i * self._I4s)

    def calc_C_eff(self):
        self.A_MT_i = self.calc_A_MT_i(c_i=self._c_i, A_SI_i=self.calc_A_SI_i())
        self.C_eff = self.calc_C_eff_by_A_i(
            c_i=self._c_i, A_i=self.A_MT_i, C_i=self._C_i, C_m=self._C_m
        )
        return self.C_eff

    def calc_A_SI_i(self):
        r"""Wrap
        :any:`mechmean.approximation.TwoPhaseComposite.calc_A_SI_from_hill_polarization`
        """
        return self.calc_A_SI_from_hill_polarization(
            P_i=self._P_i, C_i=self._C_i, C_m=self._C_m
        )


class MoriTanakaOrientationAveragedBenveniste(MoriTanaka):
    """See approximation_docstrings for docstrings"""

    def __init__(self, **kwargs):

        self.averaging_func = kwargs["averaging_func"]
        super().__init__(**kwargs)
        self._dC = self._C_i - self._C_m

    def calc_A_SI_i(self):

        A_SI = self.calc_A_SI_from_hill_polarization(
            P_i=self._P_i, C_i=self._C_i, C_m=self._C_m
        )
        A_star = self._dC @ A_SI
        A_star_averaged = self.averaging_func(A_star)
        return inv(self._dC) @ A_star_averaged


class Kehrer2019(object):
    """See approximation_docstrings for docstrings"""

    def __init__(self, phases, averaging_func, k):
        self.phases = phases
        self.averaging_func = averaging_func
        self.k = k

        self.Pud_func = mechmean.hill_polarization.Castaneda().needle

        self.P0_func = functools.partial(
            mechmean.hill_polarization.Castaneda().spheroid, aspect_ratio=1.0
        )

        self._I4s = mechkit.notation.Converter().to_mandel6(mechkit.tensors.Basic().I4s)

    def calc_C_eff(self):
        av = self.averaging_func
        Is = self._I4s

        k = self.k

        mat_i = self.phases["inclusion"]["material"]
        mat_m = self.phases["matrix"]["material"]

        Ci = mat_i.stiffness_mandel6
        Cm = mat_m.stiffness_mandel6
        dC = Cm - Ci

        ci = self.phases["inclusion"]["volume_fraction"]
        cm = 1.0 - ci

        #############################################################
        # Step 1

        P_ud_i = self.Pud_func(matrix=mat_i)
        P_ud_m = self.Pud_func(matrix=mat_m)

        tmp_upp = inv(Is + ci * P_ud_i @ dC)
        tmp_low = inv(Is - cm * P_ud_m @ dC)

        C_upp = Ci + cm * dC @ tmp_upp
        C_low = Cm - ci * dC @ tmp_low

        self._C_ud_upp = C_upp
        self._C_ud_low = C_low

        #############################################################
        # Step 2

        ref_mat = (1.0 - k) * mat_m + k * mat_i

        self._C0 = C0 = ref_mat.stiffness_mandel6

        P0 = self.P0_func(matrix=ref_mat)
        P0inv = inv(P0)

        Astar_upp = inv(P0inv + C_upp - C0)
        Astar_low = inv(P0inv + C_low - C0)

        self.C_eff_upp = C0 - P0inv + inv(av(Astar_upp))
        self.C_eff_low = C0 - P0inv + inv(av(Astar_low))

        NT = namedtuple("Effective_stiffness", ["upper", "lower"])
        return NT(self.C_eff_upp, self.C_eff_low)


class HashinShtrikmanWalpole(object):
    def __init__(self, phases, P_func, averaging_func):
        self.P_func = P_func
        for key, phase in phases.items():
            phases[key]["stiffness"] = phase["material"].stiffness_mandel6
        self.phases = phases
        self.averaging_func = averaging_func

    def calc_C_eff(self, ref_material):
        P = self._calc_P(ref_material=ref_material)

        func = self._select_func_depending_on_P(P)

        return func(
            phases=copy.deepcopy(self.phases), P=P, C_0=ref_material.stiffness_mandel6
        )

    def _calc_P(self, ref_material):
        return self.P_func(matrix=ref_material)

    def _select_func_depending_on_P(self, P):
        invertible = mechmean.utils.isinvertible(P)
        funcs = {
            True: self._calc_C_eff_for_invertible_P,
            False: self._calc_C_eff_for_singular_P_using_dual_scheme,
        }
        return funcs[invertible]

    def _calc_C_eff_for_invertible_P(self, phases, P, C_0):
        P_inv = inv(P)

        for key, phase in phases.items():
            phases[key]["W"] = inv(phase["stiffness"] - C_0 + P_inv)

        W_av = self.averaging_func(phases=phases)

        return C_0 - P_inv + inv(W_av)

    def _calc_C_eff_for_singular_P_using_dual_scheme(self, phases, P, C_0):
        r"""Handle singular P using dual scheme following
        [Walpole1969]_ (page 238).
        """
        # Calc inverse Polarization and compliance
        con = mechkit.notation.Converter()
        I4s = con.to_mandel6(mechkit.tensors.Basic().I4s)
        Q = C_0 @ (I4s - P @ C_0)

        S_0 = inv(C_0)

        # Invert stiffnesses of phases
        for key, phase in phases.items():
            phases[key]["stiffness"] = inv(phase["stiffness"])

        S_eff = self._calc_C_eff_for_invertible_P(phases=phases, P=Q, C_0=S_0)
        return inv(S_eff)

    def calc_strain_localization_tensor(self, key_phase, ref_material):
        P = self._calc_P(ref_material=ref_material)
        C_0 = ref_material.stiffness_mandel6
        phases = copy.deepcopy(self.phases)

        func_A_SI = TwoPhaseComposite().calc_A_SI_from_hill_polarization

        for key, phase in phases.items():
            phases[key]["A_SI"] = func_A_SI(P_i=P, C_i=phase["stiffness"], C_m=C_0)

        A_SI_average = self.averaging_func(phases=phases, key_to_be_averaged="A_SI")

        return phases[key_phase]["A_SI"] @ inv(A_SI_average)

    def calc_stress_localization_tensor(self, key_phase, ref_material):
        kwargs = locals()
        A = self.calc_strain_localization_tensor(**kwargs)
        S_0 = inv(ref_material.stiffness_mandel6)
        return self.phases[key_phase]["stiffness"] @ A @ S_0

    def localize_strain(self, key_phase, ref_material, strain_macro):
        kwargs = locals()
        A = self.calc_strain_localization_tensor(**kwargs)
        return A @ strain_macro

    def localize_stress(self, key_phase, ref_material, stress_macro):
        kwargs = locals()
        B = self.calc_stress_localization_tensor(**kwargs)
        return B @ stress_macro


class HSW_VolumeFraction(HashinShtrikmanWalpole):
    """See approximation_docstrings for docstrings"""

    class AveragerKeywordWeightedSum(object):
        def __call__(
            self, phases, key_to_be_averaged="W", key_weight="volume_fraction"
        ):
            sum = 0.0
            for _, phase in phases.items():
                sum += phase[key_weight] * phase[key_to_be_averaged]
            return sum

    def __init__(self, phases, P_func):
        super().__init__(
            phases=phases,
            P_func=P_func,
            averaging_func=self.AveragerKeywordWeightedSum(),
        )


class HSW_SinglePhaseOrientation(HashinShtrikmanWalpole):
    """See approximation_docstrings for docstrings"""

    class AveragerOrientationWithKeyword(object):
        def __init__(self, averager):
            self.averager = averager

        def __call__(self, phases, key_to_be_averaged="W"):
            if len(phases) != 1:
                raise utils.Ex(
                    "More than one phase."
                    "Please use HashinShtrikmanWalpole instead of"
                    "HashinShtrikmanWalpole_SinglePhaseOrientation"
                )

            key_phase = list(phases.keys())[0]
            return self.averager(phases[key_phase][key_to_be_averaged])

    def __init__(self, phases, P_func, orientation_averager):
        super().__init__(
            phases=phases,
            P_func=P_func,
            averaging_func=self.AveragerOrientationWithKeyword(
                averager=orientation_averager,
            ),
        )


class Kehrer2019HSW(object):
    """See approximation_docstrings for docstrings"""

    def __init__(self, phases, averaging_func, k):
        self.inclusion = phases["inclusion"]["material"]
        self.volume_fraction = phases["inclusion"]["volume_fraction"]
        self.matrix = phases["matrix"]["material"]
        self.averaging_func = averaging_func
        self.k = k

        self.Pud_func = mechmean.hill_polarization.Castaneda().needle

        self.P0_func = functools.partial(
            mechmean.hill_polarization.Castaneda().spheroid, aspect_ratio=1.0
        )

    def calc_C_eff(self):

        # Step 1
        hashin1 = HSW_VolumeFraction(
            phases={
                "i": {
                    "material": self.inclusion,
                    "volume_fraction": self.volume_fraction,
                },
                "m": {
                    "material": self.matrix,
                    "volume_fraction": 1.0 - self.volume_fraction,
                },
            },
            P_func=self.Pud_func,
        )

        self.C_ud_upp = hashin1.calc_C_eff(ref_material=self.inclusion)
        self.C_ud_low = hashin1.calc_C_eff(ref_material=self.matrix)

        # Step 2
        hashin2_upp = self._get_hashin2(stiffness=self.C_ud_upp)
        hashin2_low = self._get_hashin2(stiffness=self.C_ud_low)

        k = self.k

        ref_mat = (1.0 - k) * self.matrix + k * self.inclusion

        self.C_eff_upp = hashin2_upp.calc_C_eff(ref_material=ref_mat)
        self.C_eff_low = hashin2_low.calc_C_eff(ref_material=ref_mat)

        NT = namedtuple("Effective_stiffness", ["upper", "lower"])
        return NT(self.C_eff_upp, self.C_eff_low)

    def _get_hashin2(self, stiffness):
        # ToDo: Fix this Quickfix
        class Bunch(object):
            pass

        instance = Bunch()
        instance.stiffness_mandel6 = stiffness

        return HSW_SinglePhaseOrientation(
            phases={
                "UD_material": {"material": instance},
            },
            P_func=self.P0_func,
            orientation_averager=self.averaging_func,
        )


class HSW2StepInterpolatedReferenceMaterial:
    """See approximation_docstrings for docstrings"""

    def __init__(self, phases, averaging_func, k1, k2):
        self.inclusion = phases["inclusion"]["material"]
        self.volume_fraction = phases["inclusion"]["volume_fraction"]
        self.matrix = phases["matrix"]["material"]
        self.averaging_func = averaging_func
        self.k1 = k1
        self.k2 = k2

        self.Pud_func = mechmean.hill_polarization.Castaneda().needle

        self.P0_func = functools.partial(
            mechmean.hill_polarization.Castaneda().spheroid, aspect_ratio=1.0
        )

    def calc_C_eff(self):

        # Step 1
        hashin_1 = HSW_VolumeFraction(
            phases={
                "i": {
                    "material": self.inclusion,
                    "volume_fraction": self.volume_fraction,
                },
                "m": {
                    "material": self.matrix,
                    "volume_fraction": 1.0 - self.volume_fraction,
                },
            },
            P_func=self.Pud_func,
        )

        self.C_ud = hashin_1.calc_C_eff(
            ref_material=self.interpolate_reference_material(k=self.k1)
        )

        # Step 2
        hashin2 = HSW_SinglePhaseOrientation(
            phases={
                "UD_material": {
                    "material": self.make_stiffness_attribute_of_object(
                        stiffness=self.C_ud
                    ),
                },
            },
            P_func=self.P0_func,
            orientation_averager=self.averaging_func,
        )

        self.C_eff = hashin2.calc_C_eff(
            ref_material=self.interpolate_reference_material(k=self.k2)
        )

        return self.C_eff

    def interpolate_reference_material(self, k):
        return (1.0 - k) * self.matrix + k * self.inclusion

    def make_stiffness_attribute_of_object(self, stiffness):
        # ToDo: Fix this Quickfix
        class Bunch(object):
            pass

        instance = Bunch()
        instance.stiffness_mandel6 = stiffness

        return instance





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Orientation averager
"""

import numpy as np
import mechkit
import mechmean


class AdvaniTucker:
    r"""Orientation averaging following [Advani1987]_

    The orientation average of a fourth order tensor
    :math:`\left<\mathbb{A}\right>_{\text{orientation}}`
    is calculated based on fabric tensors of the first kind
    of order two
    :math:`\boldsymbol{N}`
    and four
    :math:`\mathbb{N}`.

    If the symmetry axis of :math:`\mathbb{A}` is aligned with the base
    direction of the first tensor index
    :math:`\boldsymbol{e}_{0}`,
    the average is given by

        .. math::
            \begin{align}
                \left<\mathbb{A}\right>_{\text{orientation}}
                &=
                b_0\mathbb{N}             \\
                &+
                b_1 \left(
                        \boldsymbol{N} \otimes \boldsymbol{I}
                    +   \boldsymbol{I} \otimes \boldsymbol{N}
                    \right)         \\
                &+
                b_2 \left(
                        \boldsymbol{N} \Box \boldsymbol{I}
                    +   \left( \boldsymbol{N} \Box \boldsymbol{I} \right)^{T_\text{R}}
                    +   \left( \boldsymbol{I} \Box \boldsymbol{N} \right)^{T_\text{H}}
                    +   \left( \boldsymbol{I} \Box \boldsymbol{N} \right)^{T_\text{R}}
                    \right)         \\
                &+
                b_3 \boldsymbol{I} \otimes \boldsymbol{I} \\
                &+
                b_4 \mathbb{I}^{\text{S}}
            \end{align}

    with coefficients

        .. math::
            \begin{align}
                b_0     &= A_{0000} + A_{1111} - 2A_{0011} - 4A_{0101}                  \\
                b_1     &= A_{0011} - A_{1122}                                          \\
                b_2     &= A_{0101} + \frac{1}{2} \left( A_{1122} - A_{1111} \right)    \\
                b_3     &= A_{1122}                                                     \\
                b_4     &= A_{1111} - A_{1122}
            \end{align}

    References
    ----------

    .. [Advani1987] Advani, S.G. and Tucker III, C.L., 1987.
        The use of tensors to describe and predict fiber orientation in short
        fiber composites. Journal of rheology, 31(8), pp.751-784.

    """

    def __init__(self, N4):
        self._con = mechkit.notation.Converter()
        self.base = self.get_base(N4)

    def get_base(self, N4):
        """Calc bases of factors :math:`b_i`"""
        tensors = mechkit.tensors.Basic()
        I2 = tensors.I2
        I4s = tensors.I4s

        N4 = self._con.to_tensor(N4)
        N2 = np.einsum("ijkl, kl->ij", N4, I2)

        def dyad(A, B):
            return np.einsum("ij, kl->ijkl", A, B)

        def box(A, B):
            return np.einsum("ij, kl->iklj", A, B)

        N2_box_I2 = box(N2, I2)
        I2_box_N2 = box(I2, N2)

        base = np.zeros((5, 3, 3, 3, 3))
        base[0, :] = N4

        base[1, :] = dyad(N2, I2) + dyad(I2, N2)

        base[2, :] = (
            N2_box_I2
            + np.einsum("ijkl->ijlk", N2_box_I2)
            + np.einsum("ijkl->klij", I2_box_N2)
            + np.einsum("ijkl->ijlk", I2_box_N2)
        )
        base[3, :] = dyad(I2, I2)

        base[4, :] = I4s

        return base

    def average(self, B):
        """Calc average of B"""

        # TODO: Average both B4 and B2
        # TODO: Ask for N-tensor of same order as B-tensor
        #       and calc lower orders of N by contraction

        A = self._con.to_tensor(B)

        b = np.zeros((5))
        b[0] = A[0, 0, 0, 0] + A[1, 1, 1, 1] - 2.0 * A[0, 0, 1, 1] - 4.0 * A[0, 1, 0, 1]
        b[1] = A[0, 0, 1, 1] - A[1, 1, 2, 2]
        b[2] = A[0, 1, 0, 1] + 0.5 * (A[1, 1, 2, 2] - A[1, 1, 1, 1])
        b[3] = A[1, 1, 2, 2]
        b[4] = A[1, 1, 1, 1] - A[1, 1, 2, 2]

        av = np.einsum("m, mijkl->ijkl", b, self.base)

        return self._con.to_mandel6(av)



import setuptools

setuptools.setup(
    name="mechmean",
    version="0.0.2",
    author="Julian Karl Bauer",
    author_email="juliankarlbauer@gmx.de",
    description="mechmean",
    long_description="This Python package contains selected mean field methods"
    " in the context of continuum mechanics"
    " with special focus on orientation averaged homogenization",
    url="https://github.com/JulianKarlBauer/mechmean/",
    packages=["mechmean"],
    package_dir={"mechmean": "mechmean"},
    install_requires=[
        "numpy",
        "scipy",
        "mechkit>=0.2.6",
        # "natsort",
    ],
    # setup_requires=["pybind11>=2.3", "libcgal-dev", "libeigen3-dev"],
    # extras_require={"test": ["pytest", "natsort"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np
import itertools
import mechkit

import mechmean
from mechmean import hill_polarization_alternatives


np.set_printoptions(
    linewidth=140,
    precision=2,
    # suppress=False,
)


def test_use_factory():

    factory = mechmean.hill_polarization.Factory()

    K, G = (1e6, 4e5)
    matrix = mechkit.material.Isotropic(K=K, G=G)

    _ = factory.spheroid(aspect_ratio=10, matrix=matrix)
    _ = factory.sphere(matrix=matrix)
    _ = factory.needle(matrix=matrix)


def test_compare_Hill_polarization_Castaneda_Mura():

    factory = mechmean.hill_polarization.Factory()

    for K, G in [(1e6, 4e5), (1666.6, 769.3), (200, 100)]:
        matrix = mechkit.material.Isotropic(K=K, G=G)
        for aspect_ratio in [10, 1.5, 0.5, 0.1]:

            # Mura
            P_Mura = hill_polarization_alternatives.Mura().spheroid(
                aspect_ratio=aspect_ratio, matrix=matrix
            )

            # Castaneda
            P_Casta = factory.spheroid(aspect_ratio=aspect_ratio, matrix=matrix)

            # Compare
            print("\nK = {}\nG = {}\naspect_ratio = {}".format(K, G, aspect_ratio))
            print("P_h == P_e is", np.allclose(P_Mura, P_Casta))

            assert np.allclose(P_Mura, P_Casta)


def test_compare_Hill_implementations():
    def get_polarization_alternative_implementation(aspect_ratio, matrix):
        Km = matrix.K
        Gm = matrix.G
        ar = aspect_ratio

        arsqu = ar ** 2.0
        denom = Gm * (1.0 - arsqu) * (4.0 * Gm + 3.0 * Km)
        if ar < 1.0:
            print("ATTENTION: aspect ratio < 1 ..formulas might be incorrect!")
            h = (
                ar
                * (np.arccos(ar) - ar * np.sqrt(1.0 - arsqu))
                * (1.0 - arsqu) ** (-1.5)
            )
        else:
            h = (
                ar
                * (ar * np.sqrt(arsqu - 1.0) - np.arccosh(ar))
                / (arsqu - 1.0) ** (1.5)
            )

        P0k = (
            Gm * (7 * h - 2 * arsqu - 4 * arsqu * h)
            + 3 * Km * (h - 2 * arsqu + 2 * arsqu * h)
        ) / (8.0 * denom)
        P0m = ((Gm + 3 * Km) * (2 * arsqu - h - 2 * arsqu * h)) / (4.0 * denom)
        P0r = (
            Gm * (6 - 5 * h - 8 * arsqu + 8 * arsqu * h)
            + 3 * Km * (h - 2 * arsqu + 2 * arsqu * h)
        ) / (2.0 * denom)
        P0p = (
            Gm * (15 * h - 2 * arsqu - 12 * arsqu * h) + 3 * Km * (3 * h - 2 * arsqu)
        ) / (16.0 * denom)
        P0q = (
            2 * Gm * (4 - 3 * h - 2 * arsqu)
            + 3 * Km * (2 - 3 * h + 2 * arsqu - 3 * arsqu * h)
        ) / (8.0 * denom)

        P0 = np.array(
            [
                [P0r, P0m, P0m, 0, 0, 0],
                [P0m, P0k + P0p, P0k - P0p, 0, 0, 0],
                [P0m, P0k - P0p, P0k + P0p, 0, 0, 0],
                [0, 0, 0, 2 * P0p, 0, 0],
                [0, 0, 0, 0, 2 * P0q, 0],
                [0, 0, 0, 0, 0, 2 * P0q],
            ]
        )
        return P0

    factory = mechmean.hill_polarization.Factory()

    for K, G in [
        (1e6, 4e5),
        (1666.0, 769.0),
        (300, 120),
    ]:
        matrix = mechkit.material.Isotropic(K=K, G=G)
        for aspect_ratio in [0.4, 1.1, 10, 100]:
            # Hill Bauer
            P_ref = factory.spheroid(aspect_ratio=aspect_ratio, matrix=matrix)
            # Hill alternative
            P_2 = get_polarization_alternative_implementation(
                aspect_ratio=aspect_ratio, matrix=matrix
            )
            assert np.allclose(P_ref, P_2)


def test_Hill_spheroid_check_tolerance_sphere_Mura():
    """As tol in this function is smaller than tol in decision-maker deciding
    between spheroid and sphere, this function ONLY calculates SPHERES."""

    tol = 1e-5

    for K, G in [(1e6, 4e5), (2000, 1000)]:
        matrix = mechkit.material.Isotropic(K=K, G=G)
        P = []
        for aspect_ratio in [1.0, 1.0 - tol, 1.0 + tol]:
            P_tmp = hill_polarization_alternatives.Mura().spheroid(
                aspect_ratio=aspect_ratio, matrix=matrix
            )
            P.append(P_tmp)
        print("K=", K, "G=", G)
        print(P)
        combinations = list(itertools.combinations(P, 2))
        assert all([np.allclose(x, y) for x, y in combinations])


def test_Hill_spheroid_check_tolerance_sphere_Castaneda():
    """As tol in this function is smaller than tol in decision-maker deciding
    between spheroid and sphere, this function ONLY calculates SPHERES."""

    tol = 1e-5

    for K, G in [(1e6, 4e5), (2000, 1000)]:
        matrix = mechkit.material.Isotropic(K=K, G=G)
        P = []
        for aspect_ratio in [1.0, 1.0 - tol, 1.0 + tol]:
            P_tmp = mechmean.hill_polarization.Castaneda().spheroid(
                aspect_ratio=aspect_ratio, matrix=matrix
            )
            P.append(P_tmp)
        print("K=", K, "G=", G)
        print(P)
        combinations = list(itertools.combinations(P, 2))
        assert all([np.allclose(x, y) for x, y in combinations])


def test_compare_Hill_Castaneda_Mura_sphere():

    aspect_ratio = 1.0
    for K, G in [(1e6, 4e5), (1666.6, 769.3), (200, 100)]:
        matrix = mechkit.material.Isotropic(K=K, G=G)

        # Mura
        P_Mura = hill_polarization_alternatives.Mura().spheroid(
            aspect_ratio=aspect_ratio, matrix=matrix
        )

        # Castaneda
        P_Casta = mechmean.hill_polarization.Castaneda().spheroid(
            aspect_ratio=aspect_ratio, matrix=matrix
        )

        # Compare
        print("\nK = {}\nG = {}\naspect_ratio = {}".format(K, G, aspect_ratio))
        print("P_Casta == P_Mura is", np.allclose(P_Casta, P_Mura))
        print(P_Casta - P_Mura)

        assert np.allclose(P_Casta, P_Mura)


def test_compare_Hill_needle_as_limit_spheroid():

    aspect_ratio = 10000.0
    for K, G in [(1e6, 4e5), (1666.6, 769.3), (200, 100)]:
        matrix = mechkit.material.Isotropic(K=K, G=G)

        # Mura
        P_Mura = hill_polarization_alternatives.Mura().spheroid(
            aspect_ratio=aspect_ratio, matrix=matrix
        )

        # Castaneda
        P_Casta = mechmean.hill_polarization.Castaneda().spheroid(
            aspect_ratio=aspect_ratio, matrix=matrix
        )

        # Needle
        P_needle = mechmean.hill_polarization.Castaneda().needle(matrix=matrix)

        # Compare
        assert np.allclose(P_needle, P_Mura)
        assert np.allclose(P_needle, P_Casta)


def test_compare_Hill_needle_Castaneda_Ortolano():

    for K, G in [(1e6, 4e5), (1666.6, 769.3), (200, 100)]:
        matrix = mechkit.material.Isotropic(K=K, G=G)

        P_Ortolano = mechmean.hill_polarization_alternatives.Ortolano().needle(
            matrix=matrix
        )

        P_Casta = mechmean.hill_polarization.Castaneda().needle(matrix=matrix)

        # Compare
        assert np.allclose(P_Casta, P_Ortolano)


if __name__ == "__main__":

    factory = mechmean.hill_polarization.Factory()

    for K, G in [(1e6, 4e5)]:
        matrix = mechkit.material.Isotropic(K=K, G=G)
        for aspect_ratio in [1.0]:

            # Mura
            P_hill = factory.spheroid(aspect_ratio=aspect_ratio, matrix=matrix)

            print(P_hill)
